from abc import ABCMeta, abstractmethod
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Generic,
    Iterator,
    Mapping,
    Sequence,
    Text,
    Tuple,
    TypeVar,
    Union,
)

if TYPE_CHECKING:
    from .ast import Final

# Tokens type
Tok = TypeVar("Tok")

# Output type
Out = TypeVar("Out")


class Matcher(Generic[Tok, Out], metaclass=ABCMeta):
    @abstractmethod
    def match(self, token: Tok) -> Iterator[Out]:
        raise NotImplementedError


class Eq(Matcher):
    def __init__(self, ref: Tok):
        self.ref = ref

    def match(self, token: Tok) -> Iterator[Out]:
        if self.ref == token:
            yield token

    def __repr__(self):
        return f"Eq({self.ref!r})"


class In(Matcher):
    def __init__(self, ref: Sequence[Tok]):
        self.ref = ref

    def match(self, token: Tok) -> Iterator[Out]:
        if token in self.ref:
            yield token

    def __repr__(self):
        return f"In({self.ref!r})"


class OutOf(Matcher):
    def __init__(self, ref: Tok):
        self.ref = ref

    def __repr__(self):
        return f"OutOf({self.ref!r})"

    def match(self, token: Tok) -> Iterator[Out]:
        if self.ref in token:
            yield self.ref


class AttributeHasValue(Matcher):
    def __init__(self, attribute: Text, value: Any):
        self.attribute = attribute
        self.value = value

    def match(self, token: Tok) -> Iterator[Out]:
        if (
            hasattr(token, self.attribute)
            and getattr(token, self.attribute) == self.value
        ):
            yield token

    def __repr__(self):
        return f"AttributeHasValue({self.attribute}={self.value!r}"


class KeyHasValue(Matcher):
    def __init__(self, key: Any, value: Any):
        self.key = key
        self.value = value

    def match(self, token: Tok) -> Iterator[Out]:
        if (
            isinstance(token, Mapping)
            and self.key in token
            and token[self.key] == self.value
        ):
            yield token

    def __repr__(self):
        return f"KeyHasValue({self.key!r}={self.value!r})"


class Anything(Matcher):
    def __repr__(self):
        return f"Anything()"

    def match(self, token: Tok) -> Iterator[Out]:
        yield token


class ChrRanges(Matcher[str, str]):
    def __init__(self, *ranges: Tuple[str, str]):
        self.ranges = ranges

    def __repr__(self):
        return f"ChrRanges{self.ranges!r}"

    def match(self, token: Tok) -> Iterator[Out]:
        for start, stop in self.ranges:
            if ord(start) <= ord(token) <= ord(stop):
                yield token


class Test(Matcher[Tok, Out]):
    """
    Runs an arbitrary test and matches the token as-is if successful
    """

    def __init__(self, test: Callable[[Tok], bool]):
        self.test = test

    def __repr__(self):
        return f"Test({self.test!r}"

    def match(self, token: Tok) -> Iterator[Out]:
        if self.test(token):
            yield token


class Not(Matcher[Tok, Out]):
    """
    Negates another matcher. If no match is found then the token is returned
    as-is.
    """

    def __init__(self, matcher: Union[Matcher, "Final"]):
        from .ast import Final

        if isinstance(matcher, Final):
            self.matcher: Matcher[Tok, Out] = matcher.statement
        elif isinstance(matcher, Matcher):
            self.matcher: Matcher[Tok, Out] = matcher
        else:
            raise ValueError("Cannot negate this")

    def __repr__(self):
        return f"Not({self.matcher!r})"

    def match(self, token: Tok) -> Iterator[Out]:
        found = list(self.matcher.match(token))

        if not found:
            yield token


__all__ = [
    "Tok",
    "Out",
    "Matcher",
    "Eq",
    "In",
    "OutOf",
    "AttributeHasValue",
    "KeyHasValue",
    "Anything",
    "ChrRanges",
    "Test",
    "Not",
]



from dataclasses import dataclass, field, replace
from functools import reduce
from typing import Generic, Text, Union

from .matchers import Matcher, Out, Tok


@dataclass(frozen=True)
class Node(Generic[Tok, Out]):
    """
    Root class for a node. It has no real usage by itself but that's useful to
    define operators.

    Assembling the nodes together will build an AST which the RegExp class will
    then turn into a compiled regular expression.

    Example:

    >>> from nsre import *
    >>> root = Final(Eq('a')) + Final(Eq('b')) * slice(1, 5)
    >>> re = RegExp.from_ast(root)
    >>> assert re.match('abb')
    >>> assert not re.match('a')
    """

    def __add__(self, other: "Node"):
        """
        Generates a concatenation
        """

        return Concatenation(self, other)

    def __or__(self, other: "Node"):
        """
        Generates an alternation
        """

        return Alternation(self, other)

    def __getitem__(self, item: Text):
        """
        Generates a capture group
        """

        if not isinstance(item, Text):
            raise KeyError("Cannot capture with a key other than a string")

        return Capture(name=item, statement=self)

    def __mul__(self, other: Union[int, slice]):
        """
        Generates a repetition

        - If just an integer, repeat exactly this number of times
        - If a slice(), repeat between start to stop times (included). If the
          values of start or stop are None then it's respectively considered
          like 0 and +inf.
        """

        if isinstance(other, int):
            if other < 1:
                raise ValueError("Cannot repeat item a negative number of times")

            return reduce(lambda a, b: a + b, [replace(self) for _ in range(0, other)])
        elif isinstance(other, slice):
            parts = []

            if isinstance(other.start, int) and other.start > 0:
                parts.append(self * other.start)
            elif other.start is None or other.start == 0:
                pass
            else:
                raise ValueError("Start of slice does not look valid")

            if isinstance(other.stop, int):
                for _ in range(other.start or 0, other.stop):
                    parts.append(Maybe(replace(self)))
            elif other.stop is None:
                parts.append(AnyNumber(replace(self)))
            else:
                raise ValueError("End of slice does not look valid")

            return reduce(lambda a, b: a + b, parts)
        else:
            raise ValueError("Multiply either with an int or a slice")

    def copy(self):
        """
        Generates a copy of the node. This is done because of the way the graph
        generation works: it will put all the nodes in a graph so all of them
        will need a unique ID in case the same sub-tree was used several cases.
        """

        return replace(self)


# noinspection PyUnresolvedReferences
class CopyLeftRightMixin:
    """
    Mixin to help with the copy of nodes that have a left and a right attribute
    """

    def copy(self):
        return self.__class__(self.left.copy(), self.right.copy())


# noinspection PyUnresolvedReferences
class CopyStatementMixin:
    """
    Mixin to help with the copy of nodes that just have a statement attribute
    """

    def copy(self):
        return self.__class__(self.statement.copy())


class DumbHash:
    """
    Mixin to generate a very dumb __hash__ and __eq__ implementation based on
    the object's ID. This simplifies the construction of the graph.
    """

    def __hash__(self):
        return hash(id(self))

    def __eq__(self, other):
        return self is other


@dataclass(frozen=True, eq=False)
class Final(DumbHash, Node):
    """
    In the end, all nodes in the graph should be Final(). They allow the
    engine to call the matcher (stored in statement here).
    """

    statement: Matcher[Tok, Out]

    def __lt__(self, other):
        """
        Comparable for use in the de-duplication process
        """

        return id(self) < id(other)


@dataclass(frozen=True, eq=False)
class Concatenation(DumbHash, CopyLeftRightMixin, Node):
    """
    Represents a concatenation of the left and the right nodes.
    """

    left: Node
    right: Node


@dataclass(frozen=True, eq=False)
class Alternation(DumbHash, CopyLeftRightMixin, Node):
    """
    Represents an alternation of the left and the right nodes
    """

    left: Node
    right: Node


@dataclass(frozen=True, eq=False)
class Maybe(DumbHash, CopyStatementMixin, Node):
    """
    Represents 0 or 1 occurrence of the statement
    """

    statement: Node


@dataclass(frozen=True, eq=False)
class AnyNumber(DumbHash, CopyStatementMixin, Node):
    """
    Represents 0 to +inf occurrences of the statement
    """

    statement: Node


@dataclass(frozen=True, eq=False)
class Capture(DumbHash, Node):
    """
    Represents a capture group around the statement
    """

    name: Text
    statement: Node = field(repr=False)

    def copy(self):
        return Capture(name=self.name, statement=self.statement.copy())

    def __lt__(self, other):
        """
        Comparable for use in the de-duplication process
        """

        return id(self) < id(other)


@dataclass(frozen=True)
class _Initial(Node):
    """
    Special initial node. Don't use it directly.
    """


@dataclass(frozen=True)
class _Terminal(Node):
    """
    Special final node. Don't use it directly.
    """


__all__ = [
    "Node",
    "Final",
    "Concatenation",
    "Alternation",
    "Maybe",
    "AnyNumber",
    "Capture",
]



from .ast import *
from .matchers import *
from .shortcuts import *

__all__ = [
    "alnum",
    "alnums",
    "alpha",
    "alphas",
    "ascii_alnum",
    "ascii_alnums",
    "ascii_alpha",
    "ascii_alphas",
    "ascii_decimal",
    "ascii_decimals",
    "ascii_digit",
    "ascii_digits",
    "ascii_lower_alnum",
    "ascii_lower_alnums",
    "ascii_lower_alpha",
    "ascii_lower_alphas",
    "ascii_numeric",
    "ascii_numerics",
    "ascii_printable",
    "ascii_printables",
    "ascii_space",
    "ascii_spaces",
    "ascii_upper_alnum",
    "ascii_upper_alnums",
    "ascii_upper_alpha",
    "ascii_upper_alphas",
    "decimal",
    "decimals",
    "digit",
    "digits",
    "lower_alnum",
    "lower_alnums",
    "lower_alpha",
    "lower_alphas",
    "numeric",
    "numerics",
    "printable",
    "printables",
    "space",
    "spaces",
    "spaces_maybe",
    "upper_alnum",
    "upper_alnums",
    "upper_alpha",
    "upper_alphas",
    "hex_digit",
    "hex_digits",
    "domain_name",
    "url",
    "email",
    "html_tag",
]


if hasattr(str, "isascii"):

    def _isascii(s: str) -> bool:
        """
        Use regular isascii test in Python 3.7+
        """

        return s.isascii()


else:

    def _isascii(s: str) -> bool:
        """
        Support for isascii test in Python 3.6
        """

        if not s:
            return False

        for c in s:
            if ord(c) > ord("\u007f"):
                return False

        return True


# Basic Python string tests
alnum = Final(Test(lambda t: t.isalnum()))
alnums = alnum * slice(1, None)
alpha = Final(Test(lambda t: t.isalpha()))
alphas = alpha * slice(1, None)
ascii_alnum = Final(Test(lambda t: t.isalnum() and _isascii(t)))
ascii_alnums = ascii_alnum * slice(1, None)
ascii_alpha = Final(Test(lambda t: t.isalpha() and _isascii(t)))
ascii_alphas = ascii_alpha * slice(1, None)
ascii_decimal = Final(Test(lambda t: t.isdecimal() and _isascii(t)))
ascii_decimals = ascii_decimal * slice(1, None)
ascii_digit = Final(Test(lambda t: t.isdigit() and _isascii(t)))
ascii_digits = ascii_digit * slice(1, None)
ascii_lower_alnum = Final(Test(lambda t: t.isalnum() and t.islower() and _isascii(t)))
ascii_lower_alnums = ascii_lower_alnum * slice(1, None)
ascii_lower_alpha = Final(Test(lambda t: t.isalpha() and t.islower() and _isascii(t)))
ascii_lower_alphas = ascii_lower_alpha * slice(1, None)
ascii_numeric = Final(Test(lambda t: t.isnumeric() and _isascii(t)))
ascii_numerics = ascii_numeric * slice(1, None)
ascii_printable = Final(Test(lambda t: t.isprintable() and _isascii(t)))
ascii_printables = ascii_printable * slice(1, None)
ascii_space = Final(Test(lambda t: t.isspace() and _isascii(t)))
ascii_spaces = ascii_space * slice(1, None)
ascii_upper_alnum = Final(Test(lambda t: t.isalnum() and t.isupper() and _isascii(t)))
ascii_upper_alnums = ascii_upper_alnum * slice(1, None)
ascii_upper_alpha = Final(Test(lambda t: t.isalpha() and t.isupper() and _isascii(t)))
ascii_upper_alphas = ascii_upper_alpha * slice(1, None)
decimal = Final(Test(lambda t: t.isdecimal()))
decimals = decimal * slice(1, None)
digit = Final(Test(lambda t: t.isdigit()))
digits = digit * slice(1, None)
lower_alnum = Final(Test(lambda t: t.isalnum() and t.islower()))
lower_alnums = lower_alnum * slice(1, None)
lower_alpha = Final(Test(lambda t: t.isalpha() and t.islower()))
lower_alphas = lower_alpha * slice(1, None)
numeric = Final(Test(lambda t: t.isnumeric()))
numerics = numeric * slice(1, None)
printable = Final(Test(lambda t: t.isprintable()))
printables = printable * slice(1, None)
space = Final(Test(lambda t: t.isspace()))
spaces = space * slice(1, None)
spaces_maybe = space * slice(0, None)
upper_alnum = Final(Test(lambda t: t.isalnum() and t.isupper()))
upper_alnums = upper_alnum * slice(1, None)
upper_alpha = Final(Test(lambda t: t.isalpha() and t.isupper()))
upper_alphas = upper_alpha * slice(1, None)


# Hex
hex_digit = Final(ChrRanges(("a", "f"), ("A", "F"), ("0", "9")))
hex_digits = hex_digit * slice(1, None)


# Domain name
_dn_letter = ascii_alnum
_dn_sub_part = _dn_letter * slice(1, None)
_dn_part = Maybe(seq("xn--")) + _dn_sub_part + AnyNumber(seq("-") + _dn_sub_part)
domain_name = (
    _dn_part + AnyNumber(seq(".") + _dn_part) + seq(".") + _dn_letter * slice(2, None)
)


# Email address
_em_part = ascii_alnums
_em_sep = Final(In(["+", "."]))
_em_user = _em_part + AnyNumber(_em_sep + _em_part)
email = _em_user["user"] + seq("@") + domain_name["domain"]


# URL
_url_proto = (seq("http") + Maybe(seq("s")))["proto"] + seq("://")
_url_enc = (ascii_alnum | (seq("%") + hex_digit + hex_digit) | seq("+")) * slice(
    1, None
)
_url_auth = _url_enc["user"] + seq(":") + _url_enc["password"]
_url_path = Maybe(_url_enc + AnyNumber(seq("/") + _url_enc))
_url_query_var = _url_enc + Maybe(seq("=") + _url_enc)
_url_query = _url_query_var + AnyNumber(seq("&") + _url_query_var)
_url_hash = _url_enc
url = (
    _url_proto
    + Maybe(_url_auth["auth"] + seq("@"))
    + domain_name["domain"]
    + Maybe(seq("/") + _url_path["path"])
    + Maybe(seq("?") + _url_query["query"])
    + Maybe(seq("#") + _url_hash["hash"])
)


# HTML tag
_tag_attr_name = ascii_alnums + AnyNumber(seq("-") + ascii_alnums)
_tag_attr_val_quote_1 = seq("'") + AnyNumber(Final(Not(In("'>"))))["value"] + seq("'")
_tag_attr_val_quote_2 = seq('"') + AnyNumber(Final(Not(In('">'))))["value"] + seq('"')
_tag_attr_val_savage = AnyNumber(Final(Test(lambda t: not t.isspace() and t != "/")))
_tag_attr_quote = (
    _tag_attr_name["name"]
    + Maybe(
        spaces_maybe
        + seq("=")
        + spaces_maybe
        + (_tag_attr_val_quote_1 | _tag_attr_val_quote_2)
    )
)["attr"]
_tag_attr_savage = (
    _tag_attr_name["name"]
    + Maybe(spaces_maybe + seq("=") + spaces_maybe + _tag_attr_val_savage)
)["attr"]
# _tag_inside = AnyNumber(
#     (_tag_attr_savage + spaces) | (_tag_attr_quote + spaces_maybe)
# ) + Maybe(_tag_attr_savage | _tag_attr_quote)
_tag_inside = AnyNumber((_tag_attr_quote | _tag_attr_savage) + spaces) + Maybe(
    _tag_attr_savage | _tag_attr_quote
)
html_tag = (
    seq("<")
    + spaces_maybe
    + _tag_attr_name["name"]
    + spaces
    + _tag_inside["attr"]
    + spaces_maybe
    + (seq("/>") | seq(">"))
)



from functools import reduce
from operator import add
from typing import Sequence

from .ast import AnyNumber, Final, Node
from .matchers import Anything, Eq


def seq(s: Sequence) -> Node:
    """
    Converts a sequence of items into concatenated Eq matchers. Useful to
    match a string exactly.

    >>> from nsre import *
    >>> re = RegExp.from_ast(seq('foo'))
    >>> assert re.match('foo')
    >>> assert not re.match('bar')

    Parameters
    ----------
    s
        Sequence to convert into AST
    """

    nodes = [Final(Eq(x)) for x in s]
    return reduce(add, nodes)


def anything() -> Node:
    """
    Matches anything. Conceptual equivalent of ".*" in regular regular
    expressions.
    """

    return AnyNumber(Final(Anything()))


__all__ = ["seq", "anything"]



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['nsre']

package_data = \
{'': ['*']}

install_requires = \
['networkx>=2.0,<3.0']

extras_require = \
{':python_version >= "3.6" and python_version < "3.7"': ['dataclasses>=0.6,<0.7']}

setup_kwargs = {
    'name': 'nsre',
    'version': '0.2.0',
    'description': 'Regular expressions for the 21st century. Match any data type.',
    'long_description': "Non-String Regular Expressions\n==============================\n\n[![Build Status](https://travis-ci.org/Xowap/nsre.svg?branch=develop)](https://travis-ci.org/Xowap/nsre)\n\nNSRE (Non-String Regular Expressions) is a new spin at regular expressions.\nIt's really abstract, even compared to regular expressions as you know them,\nbut it's also pretty powerful for some uses.\n\nHere's the twist: what if regular expressions could, instead of matching just\ncharacter strings, match any sequence of anything?\n\n```python\nfrom nsre import *\n\nre = RegExp.from_ast(seq('hello, ') + (seq('foo') | seq('bar')))\nassert re.match('hello, foo')\n```\n\nThe main goal here is matching NLU grammars when there is several possible\ninterpretations of a single word, however there is a lot of other things that\nyou could do. You just need to understand what NSRE is and apply it to\nsomething.\n\n> **Note** &mdash; This is inspired by\n> [this article](https://swtch.com/~rsc/regexp/regexp1.html) from Russ Cox,\n> which explains how Thompson NFA work, except that I needed to add some\n> features and then the implementation is much less elegant because I actually\n> don't know what I'm doing. But it seems to be working.\n\n## Documentation\n\n[✨ **Documentation is there** ✨](http://nsre.rtfd.io/)\n\n## Licence\n\nThis library is provided under the terms of the [WTFPL](./LICENSE).\n\nIf you find it useful, you can have a look at the\n[contributors](https://github.com/Xowap/nsre/graphs/contributors) page to\nknow who helped.\n",
    'author': 'Rémy Sanchez',
    'author_email': 'remy.sanchez@hyperthese.net',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'http://nsre.rtfd.io/',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
